﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace DataStructureLab
{
    public class QueryTests
    {
        /// <summary>
        /// Main Query Operations from Direct to Linq
        /// </summary>
        public static void QueryOperations()
        {
            var data = GetData();
            //Calling Direct Query
            Console.WriteLine("Filtering with Direct Code..... ");
            FilterMapReduce(data);

            //Calling with Predicate --> SampleFiletr
            Console.WriteLine("Calling with Predicate --> SampleFiletr....");
            FilterWithPredicate(data, SampleFilter);

            //Calling with Predicate --> ProfessionTest
            Console.WriteLine("Calling with Predicate --> ProfessionTest....");
            FilterWithPredicate(data, ProfesionTest);

            //Calling with Anonymous Function
            Console.WriteLine("Calling with Anonymous Function");
            FilterWithPredicate(data, delegate (Registration r)
            {
                if (r.Grade == "Second" && r.RegCount < 200)
                    return true;
                return false;
            });

            //Anonymous Function using Lampda 
            Console.WriteLine("Anonymous Function using Lampda ");
            FilterWithPredicate(data, (r) =>
            {
                return r.Grade == "Second" && r.RegCount < 200;
            });

            //Calling With Lampda Expression
            Console.WriteLine("Calling With Lampda Expression");
            FilterWithPredicate(data,
                r => r.Grade == "Second" && r.RegCount < 200
            );
            Console.WriteLine();
            //Calling With another expression --> Another Condition
            Console.WriteLine("Calling With another expression --> Another Condition");
            FilterWithPredicate(data, r => r.Profession == "Airframe");

            //Operations Using Linq
            Console.WriteLine("-- Operations Using Linq with Extention Methods");
            FilterUsingLinqWithExtentionMethods(data);

            //Operations Using Linq Query
            Console.WriteLine("-- Operations Using Linq Query");
            FilterUsingLinqWithQueryExpressions(data);
        }

        /// <summary>
        /// Do the Query Operations : Filter - Map - Reduce directly
        /// Note the Functions that we can exctract out of the Query Loop
        /// </summary>
        public static void FilterMapReduce(IEnumerable<Registration> data)
        {
            int sum = 0; 
            foreach (var r in data)
            {
                //Filter --> Where --> if(condition)
                //Condition --> Predicate --> bool Function(CurrentObject)
                if (r.Grade == "Second" && r.RegCount < 200)
                {
                    //Map --> Select --> NewObject = MapFunction(CurrentObject)
                    var msg = "We have " + r.RegCount.ToString()
                  + " Students for Profession: " + r.Profession;
                    Console.WriteLine(msg);
                    //Reduce --> Aggregate --> SingleValue = AggFunction(All Records)
                    sum = sum + r.RegCount;
                }
            }
            //Print Reduce Result
            Console.WriteLine("Total Number of Students = {0}", sum);
        }

        /// <summary>
        /// Define the Predicate Signature in a delegate object
        /// </summary>
        /// <param name="r"></param>
        /// <returns></returns>
        public delegate bool RegPredicate(Registration r);

        public static bool SampleFilter(Registration r)
        {
            return r.Grade == "Second" && r.RegCount < 200;
        }

        public static bool ProfesionTest(Registration r)
        {
            return r.Profession == "Mech";
            //if (r.Profession == "Mech") return true;
            //return false;
        }


        /// <summary>
        /// Example Of Exctracting the Condition Function
        /// </summary>
        /// <param name="condition"></param>
        public static void FilterWithPredicate(IEnumerable<Registration> data, RegPredicate condition)
        {
            foreach (var r in data)
            {
                //Filter
                if (condition(r))
                {
                    Console.WriteLine(r);
                }
            }
        }

        public static void FilterUsingLinqWithExtentionMethods(IEnumerable<Registration> data)
        {
            //Extention Methods
            var query =
                data.Where(r => r.Grade == "Second")
                .Where(r => r.RegCount < 200);

            foreach (var r in query)
            {
                Console.WriteLine(r);
            }
        }

        public static void FilterUsingLinqWithQueryExpressions(IEnumerable<Registration> data)
        {
            //Query Syntax
            var query = from r in data
                        where r.RegCount < 200
                        select r;

            foreach (var r in query)
            {
                Console.WriteLine(r);
            }
        }

        public static IEnumerable<Registration> GetData()
        {
            return new List<Registration>
            {
                new Registration { Profession="Communication", Grade="Second", RegCount = 281 },
                new Registration { Profession="Computer", Grade="First", RegCount = 71 },
                new Registration { Profession="Computer", Grade="Third", RegCount = 141 },
                new Registration { Profession="Computer", Grade="Second", RegCount = 203 },
                new Registration { Profession="Airframe", Grade="First", RegCount = 68 },
                new Registration { Profession="Airframe", Grade="Third", RegCount = 12 },
                new Registration { Profession="Airframe", Grade="Second", RegCount = 67 },
                new Registration { Profession="Electric", Grade="First", RegCount = 447 },
                new Registration { Profession="Electric", Grade="Third", RegCount = 166 },
                new Registration { Profession="Electric", Grade="Second", RegCount = 337 },
                new Registration { Profession="Mech", Grade="First", RegCount = 543 },
                new Registration { Profession="Mech", Grade="Third", RegCount = 584 },
                new Registration { Profession="Mech", Grade="Second", RegCount = 864 },
                new Registration { Profession="Torism", Grade="First", RegCount = 33 },
                new Registration { Profession="Torism", Grade="Third", RegCount = 127 },
                new Registration { Profession="Torism", Grade="Second", RegCount = 166 },
                new Registration { Profession="Civil", Grade="First", RegCount = 21 },
                new Registration { Profession="Civil", Grade="Third", RegCount = 29 },
                new Registration { Profession="Civil", Grade="Second", RegCount = 35 },
                new Registration { Profession="Structure", Grade="First", RegCount = 287 },
                new Registration { Profession="Structure", Grade="Third", RegCount = 228 },
                new Registration { Profession="Structure", Grade="Second", RegCount = 403 },
                new Registration { Profession="Communication", Grade="First", RegCount = 152 },
                new Registration { Profession="Communication", Grade="Third", RegCount = 260 },
                new Registration { Profession="Communication", Grade="Second", RegCount = 281 },
                new Registration { Profession="BIS", Grade="First", RegCount = 5 },
                new Registration { Profession="BIS", Grade="Third", RegCount = 11 },
                new Registration { Profession="BIS", Grade="Second", RegCount = 16 },
                new Registration { Profession="Media", Grade="First", RegCount = 1 },
                new Registration { Profession="Media", Grade="Third", RegCount = 13 },
                new Registration { Profession="Media", Grade="Second", RegCount = 11 },
                new Registration { Profession="CommerceE", Grade="First", RegCount = 541 },
                new Registration { Profession="CommerceE", Grade="Third", RegCount = 652 },
                new Registration { Profession="CommerceE", Grade="Second", RegCount = 740 },
                new Registration { Profession="CommerceA", Grade="First", RegCount = 71 },
                new Registration { Profession="CommerceA", Grade="Third", RegCount = 340 },
                new Registration { Profession="CommerceA", Grade="Second", RegCount = 266 },
                new Registration { Profession="Law", Grade="First", RegCount = 16 },
                new Registration { Profession="Law", Grade="Third", RegCount = 32 },
                new Registration { Profession="Law", Grade="Second", RegCount = 32 },
            };

        }
    }
}
